#pragma once

#define LEOHELPERS_FILE_INCLUDED

namespace LeoHelpers
{
	// Some helpers take an object which lets you redirect their Win32 API calls to alternatives/wrappers (e.g. for UAC support and/or IStreams).
	// The FileOpFuncs_Direct class here provides a default which just calls the real Win32 APIs directly.
	class FileOpFuncs_Direct
	{
	public:
		// Users of this class assume its destructor will not change the last-error value. (May not be true for alternative classes.)
		// This class and all alternatives must use SetLastError. The SetLastError calls are implicit here (the Win32 APIs do them
		// for us), but when making an altnernative class remember to preserve and set the last-error values appropriately.

		// Rules for other FileOpFuncs pseudo-interface implementations:
		// - These functions must SetLastError on failure.
		// - These functions must match the signatures of those in LeoHelpers::FileOpFuncs_Direct.
		// - FileOpFuncs is not an actual interface; we may be passed to templated functions in LeoHelpers (or elsewhere) which expect some of these signatures.
		// - FileOpFuncs implementations can be partial. A compile-time error will occur if a missing function is actually required.

		typedef HANDLE FileHandleType; // All FileOpFuncs implementations should have this template so that external users can refer to it.

		BOOL CopyFile(LPCWSTR lpExistingFileName, LPCWSTR lpNewFileName, BOOL bFailIfExists);
		BOOL MoveFileEx(LPCWSTR lpExistingFileName, LPCWSTR lpNewFileName, DWORD dwFlags);
		BOOL DeleteFile(LPCWSTR lpFileName);
		BOOL ReplaceFile(LPCWSTR lpReplacedFileName, LPCWSTR lpReplacementFileName, LPCWSTR lpBackupFileName, DWORD dwReplaceFlags);
		BOOL CreateDirectory(LPCWSTR lpPathName);
		BOOL RemoveDirectory(LPCWSTR lpPathName);
		BOOL SetFileAttributes(LPCWSTR lpFileName, DWORD dwFileAttributes);
		BOOL SetFileTime(FileHandleType hFile, const FILETIME *pCreationTime, const FILETIME *pLastAccessTime, const FILETIME *pLastWriteTime);

		FileHandleType BadFileValue() { return INVALID_HANDLE_VALUE; }
		BOOL IsBadFileValue(FileHandleType hFile) { return (INVALID_HANDLE_VALUE == hFile); }
		BOOL IsGoodFileValue(FileHandleType hFile) { return (INVALID_HANDLE_VALUE != hFile); }
		BOOL IsSameFileValue(FileHandleType hFile1, FileHandleType hFile2) { return hFile1 == hFile2; }
		BOOL IsDifferentFileValue(FileHandleType hFile1, FileHandleType hFile2) { return hFile1 != hFile2; }

		// Callers should be aware that alternative FileOpFuncs objects may return things which are not Win32 HANDLE (or not HANDLEs at all).
		// Callers should be templated on the file-handle type so that alternative FileOpFunc objects can return whatever type makes sense for their handles.
		// Callers must not assume INVALID_HANDLE_VALUE is used for invalid handles/pointers/whatevers;
		//   instead, use the BOOL result or the IsBadFilePointer method.
		// Callers may assume that the file-handle type has valid copy-constructor and assignment-operator.
		// Implementors must ensure that on return the result==IsGoodFileValue(*pHandle), since several code-paths check the handle and ignore the result.
		BOOL OpenFile(FileHandleType *pHandle, LPCWSTR lpFileName, DWORD dwDesiredAccess, DWORD dwShareMode, DWORD dwCreationDisposition, DWORD dwFlagsAndAttributes);
		BOOL CloseFile(FileHandleType hFile);
		BOOL SetFilePointerEx(FileHandleType hFile, LARGE_INTEGER liDistanceToMove, PLARGE_INTEGER lpNewFilePointer, DWORD dwMoveMethod);
		BOOL GetFileSizeEx(FileHandleType hFile, PLARGE_INTEGER lpFileSize);
		BOOL SetEndOfFile(FileHandleType hFile);
		BOOL WriteFile(FileHandleType hFile, LPCVOID lpBuffer, DWORD nNumberOfBytesToWrite, LPDWORD lpNumberOfBytesWritten);
		BOOL FlushFileBuffers(FileHandleType hFile);

		// This must not modify the last-error value.
		void RegisterThreadElevationCancelTracking(BOOL bRegister, BOOL * pbHasCancelled);
	};

	template< class FileOpFuncs >
	class FOFElevTrackScoper
	{
	private:
		FileOpFuncs & m_funcs;
		BOOL m_bHasCancelled;
	private:
		FOFElevTrackScoper(const FOFElevTrackScoper &rhs); // disallow
		FOFElevTrackScoper &operator=(const FOFElevTrackScoper &rhs); // dissallow
	public:
		FOFElevTrackScoper(FileOpFuncs & funcs) : m_funcs(funcs), m_bHasCancelled(FALSE) { m_funcs.RegisterThreadElevationCancelTracking(TRUE,  &m_bHasCancelled); }
		~FOFElevTrackScoper() /* not virtual */                                          { m_funcs.RegisterThreadElevationCancelTracking(FALSE, &m_bHasCancelled); }
		BOOL HasFailedDueToElevationCancel()                                             { return m_bHasCancelled; }
	};

	// Wraps calls to ::GetFileAttributesEx so that directory sizes are consistently reported as zero.
	// I ran into a directory whose size was reported as 0x1000 by GetFileAttributesEx which in turn broke 3rd party code I passed the data to.
	// (I have no idea why that folder was special, nor why the 3rd party code acted on a directory size, but it sure wasted several hours tracking it down!)
	BOOL GetFileAttributesEx(LPCWSTR lpFileName, GET_FILEEX_INFO_LEVELS fInfoLevelId, LPVOID lpFileInformation);

#ifndef DOPUS_SMARTPATH_REDEFINITIONS_IN_USE
	// Wraps calls to ::GetFileAttributes to use SmartPath stuff.
	DWORD GetFileAttributes(LPCWSTR lpFileName);
	// Wraps calls to ::DeleteFile to use SmartPath stuff.
	BOOL DeleteFile(LPCWSTR lpFileName);
	// Wraps calls to ::RemoveDirectory to use SmartPath stuff.
	BOOL RemoveDirectory(LPCWSTR lpPathName);
#endif // DOPUS_SMARTPATH_REDEFINITIONS_IN_USE

	bool GetShortFilePath(std::wstring *pstrResult, const wchar_t *szInput);

	bool UrlFromPath(std::wstring *pstrResult, const wchar_t *szPath);

	bool UrlFromPathShortIfPossible(std::wstring *pstrResult, const wchar_t *szPath, bool bOnlyTryShortIfNonAscii);

//	void EncodeUri(std::wstring *pstrResult, const wchar_t *szInput, bool bAppend, bool bConvertColon, bool bProperUTF8Method);

	inline void TrimTrailingSlashes(std::wstring &strPath)
	{
		std::wstring::size_type len = strPath.length();

		while(len > 0 && (strPath[len - 1] == L'\\' || strPath[len - 1] == L'/'))
		{
			--len;
			strPath.erase(len);
		}
	}

	// A path with no slashes or the empty string returns an empty string.
	void GetParentPathString(std::wstring *pstrResult, bool bIncludeSlashIfDriveRoot);
	std::wstring *GetParentPathString(std::wstring *pstrResult, const wchar_t *path, bool bIncludeSlashIfDriveRoot);

	std::wstring *AppendPathString(std::wstring *pstrPath, const wchar_t *szAppendage);

	const wchar_t *GetLastPathPart(const wchar_t *szIn);
	inline wchar_t *GetLastPathPart(wchar_t *szIn) { return const_cast< wchar_t * >( GetLastPathPart(const_cast< const wchar_t * >( szIn )) ); }

	const wchar_t *GetExtensionPart(bool bIncludeDot, const wchar_t *szIn, const wchar_t *szEndPos);

	inline const wchar_t *GetExtensionPart(bool bIncludeDot, const wchar_t *szIn)
	{
		if (NULL == szIn)
		{
			return NULL;
		}

		return GetExtensionPart(bIncludeDot, szIn, szIn + wcslen(szIn));
	}

	inline const wchar_t *GetExtensionPart(bool bIncludeDot, const std::wstring &strIn)
	{
		std::wstring::size_type dotPos = strIn.find_last_of(L".");
		std::wstring::size_type sepPos = strIn.find_last_of(L"\\/:");

		if (dotPos == 0 // Don't consider filenames like ".hidden" to have an extension.
		||	dotPos == std::wstring::npos
		||	(sepPos != std::wstring::npos && dotPos <= (sepPos + 1)) // The last dot is before, or immediately after, the last path separator
		||	dotPos == (strIn.length() - 1)) // Don't consider "." by itself to be an extension.
		{
			return NULL;
		}

		return strIn.c_str() + dotPos + (bIncludeDot ? 0 : 1);
	}

	template < typename str > bool TestExtensionPart(str NameOrPath, const wchar_t *szExt)
	{
		if (szExt && *szExt == L'.') { ++szExt; }
		bool fWantNoExt = (!szExt || !*szExt);

		const wchar_t *szExtFound = GetExtensionPart(false, NameOrPath);
		if (!szExtFound) { return fWantNoExt; }
		if (fWantNoExt) { return false; }
		return (_wcsicmp(szExtFound, szExt) == 0);
	}

	inline void TrimExtensionPart(bool bTrimDot, std::wstring *pStr)
	{
		std::wstring::size_type dotPos = pStr->find_last_of(L".");
		std::wstring::size_type sepPos = pStr->find_last_of(L"\\/:");

		if (dotPos == 0 // Don't consider filenames like ".hidden" to have an extension.
		||	dotPos == std::wstring::npos
		||	(sepPos != std::wstring::npos && dotPos <= (sepPos + 1)) // The last dot is before, or immediately after, the last path separator
		||	dotPos == (pStr->length() - 1)) // Don't consider "." by itself to be an extension.
		{
			return;
		}

		pStr->erase(dotPos + (bTrimDot ? 0 : 1));
	}

	// For code that makes file *paths* legal, Opus7Zip has a function which could be made more generic and
	// pulled into here if needed in other places.
	inline bool IsLegalPathChar(const wchar_t &c, bool bAllowPathSeps, bool bAllowDots, bool bAllowColons)
	{
		switch(c)
		{
		default:
			return true;
		case L'\\':
		case L'/':
			return bAllowPathSeps;
		case L'.':
			return bAllowDots;
		case L':':
			return bAllowColons;
		case L'*':
		case L'?':
		case L'"':
		case L'<':
		case L'>':
		case L'|':
		case L'\0':
		case L'\r':
		case L'\n':
			return false;
		}
	}

	// Handles things like "X:" "X:\" "\\?\X:" "\\?\X:\"
	// Does not handle relative paths, or stuff like "X:\."
	// If bAllowSubDirs true then anything after the "X:\" part is fine (but ignored).
	// If bAllowSubDirs false then the input string must end after the "X:" or the "X:\".
	// *pCharOutDriveLetter (which is a char not a string) will be set to the drive letter, unless the pointer is NULL.
	// Remember that "X:" doesn't literally mean the root to many APIs
	// (it means the current dir on drive X, not neccessarily the root),
	// even though it does to most of the code in here.
	bool IsDriveRootPath(const wchar_t *szPath, bool bAllowSubDirs, wchar_t *pCharOutDriveLetter);

	// If the path is like "X:" turns it into "X:\"
	void EnsureSlashAtEndIfDriveRootPath(std::wstring *pstrPath);

	bool IsUNCPath(const wchar_t *szPath, bool *pbHasShare, bool *pbHasSubDirs);

	bool IsDriveOrUNCRoot(const wchar_t *szPath);

	// If strPath starts with "@:\" the @ will be replaced with the drive letter that the current process's exe is on.
	// This can be useful on portable (USB) installs to allow people to specify paths to other tools relative to the same drive root.
	// Returns boolean success. DOES NOT SET LASTERROR.
	bool ConvertFromAtDriveLetterPath(std::wstring &strPath);

	// If strPath starts with the same drive letter as the one the current process's exe is on, replaces that drive letter with "@:\".
	bool ConvertToAtDriveLetterPath(std::wstring &strPath);

	// Use GetLastError on failure.
	// Returns true if the directory already exists.
	// pvecCreatedDirs may be NULL if you don't care, else points to a vector that'll have any created dirs added to it in order of creation.
	// *pvecCreatedDirs will NOT be cleared and may be added to even if the overall operation fails (e.g. if some dirs could be created but not all).
	template< class FileOpFuncs >
	bool CreateDirectoryAndParents(FileOpFuncs &funcs, const wchar_t *szPath, std::vector< std::wstring > *pvecCreatedDirs)
	{
		if (szPath == NULL || *szPath == L'\0')
		{
			::SetLastError(ERROR_INVALID_NAME);
			return false;
		}

		if (funcs.CreateDirectory(szPath))
		{
			if (pvecCreatedDirs != NULL)
			{
				pvecCreatedDirs->push_back(szPath);
			}
			return true;
		}

		DWORD dwErr = ::GetLastError();

		if (dwErr == ERROR_ALREADY_EXISTS || dwErr == ERROR_FILE_EXISTS)
		{
			DWORD dwDirAttrs = GetFileAttributes(szPath); // Should be diverted to either SmartPath::_GetFileAttributes or LeoHelpers::GetFileAttributes.
			if (dwDirAttrs == INVALID_FILE_ATTRIBUTES || !(dwDirAttrs&FILE_ATTRIBUTE_DIRECTORY))
			{
				::SetLastError(dwErr);
				return false;
			}

			return true;
		}

		if (dwErr == ERROR_ACCESS_DENIED)
		{
			// Special handling for ::CreateDirectory's stupid behaviour with drive roots. 
			// It returns ERROR_ACCESS_DENIED for them if they exist.
			// (Instead of ERROR_ALREADY_EXISTS like a normal folder or network share.)
			// 23/Feb/2012: Turns out some network share roots return access-denied as well, so we also check that.
			if (LeoHelpers::IsDriveOrUNCRoot(szPath))
			{
				return true;
			}
			::SetLastError(dwErr); // In case IsDriveOrUNCRoot changed it.
		}

		// 12/Dec/18: Most things return ERROR_PATH_NOT_FOUND but Samba on OS X returns ERROR_FILE_NOT_FOUND.
		if (dwErr != ERROR_PATH_NOT_FOUND && dwErr != ERROR_FILE_NOT_FOUND)
		{
			return false;
		}

		std::wstring strParent;
		LeoHelpers::GetParentPathString(&strParent, szPath, true);

		if (strParent.empty())
		{
			::SetLastError(dwErr); // In case GetParentPathString changed it.
			return false;
		}

		// Recursive call.
		if (!LeoHelpers::CreateDirectoryAndParents(funcs, strParent.c_str(), pvecCreatedDirs))
		{
			return false;
		}

		if (funcs.CreateDirectory(szPath))
		{
			if (pvecCreatedDirs != NULL)
			{
				pvecCreatedDirs->push_back(szPath);
			}
			return true;
		}

		return false;
	}

	// Use GetLastError on failure.
	// pvecCreatedDirs may be NULL if you don't care, else points to a vector that'll have any created dirs added to it in order of creation.
	// *pvecCreatedDirs will NOT be cleared and may be added to even if the overall operation fails (e.g. if some dirs could be created but not all).
	inline bool CreateDirectoryAndParents(const wchar_t *szPath, std::vector< std::wstring > *pvecCreatedDirs)
	{
		FileOpFuncs_Direct fof_direct;
		return LeoHelpers::CreateDirectoryAndParents(fof_direct, szPath, pvecCreatedDirs);
	}

	// Should be correct for the folder even if the path contains soft-links etc.
	UINT GetDriveTypeOfFolder(const wchar_t *szPath);

	std::wstring GenerateSafeName(const wchar_t *szInput);

	bool GetTempPath(std::wstring *pstrTempPath);

	// If bDontReallyOpen is true then the function won't really create a file and always return a bad file pointer, but will still set *pstrFilepathOut to a name that is similar to what you'd get normally.
	// On failure *pstrFilepathOut is guaranteed to be empty.
	// Of course, if you don't create the file then there is nothing to stop someone else (e.g. another caller of this function)
	// creating it unexpectedly, so the name should only be considered as an example (e.g. to see if we'll generate a path with any non-ASCII chars in it) and never really used.
	// Set bTidyMode when the tempfile is visible to the user (not hidden in a temp dir) and unlikely to clash with an existing file;
	// in tidy-mode the name will be tried as-is (prefix + extension of szFilenameIn, or ".tmp") and if that already exists a counter is added.
	// If bReadWrite is true, the file is opened with GENERIC_READ|GENERIC_WRITE; otherwise it is only opened with GENERIC_WRITE.
	template< class FileOpFuncs >
	bool OpenTempFileNamePreserveExtension(FileOpFuncs &funcs, typename FileOpFuncs::FileHandleType *pFileHandle, const wchar_t *szTempPath, const wchar_t *szPrefix, const wchar_t *szFilenameOrDotExtIn, std::wstring *pstrFilepathOut, bool bDontReallyOpen, DWORD dwShareMode, DWORD dwFlagsAndAttributes, bool bTidyMode, bool bReadWrite)
	{
		bool bOpenResult = false;
		*pFileHandle = funcs.BadFileValue();

		if (NULL != pstrFilepathOut)
		{
			pstrFilepathOut->clear();
		}

		std::wstring strFileStart;

		if (szTempPath != NULL)
		{
			strFileStart = szTempPath;
		}
		else if (!LeoHelpers::GetTempPath(&strFileStart))
		{
			return false; // last error already set by GetTempPath.
		}

		if (strFileStart.length() > 0 && L'\\' != *(strFileStart.rbegin()) && L'/' != *(strFileStart.rbegin()))
		{
			strFileStart += L'\\';
		}

		strFileStart += szPrefix;

		std::wstring strExtension = L".tmp";

		if (NULL != szFilenameOrDotExtIn)
		{
			const wchar_t *szExt = wcsrchr(szFilenameOrDotExtIn, L'.');

			if (NULL != szExt)
			{
				strExtension = szExt;
			}
		}

		SYSTEMTIME systime;
		GetSystemTime(&systime);

		int i1 = systime.wYear;
		int i2 = systime.wMonth;
		int i3 = systime.wDay;
		int i4 = systime.wHour;
		int i5 = systime.wMinute;
		int i6 = systime.wSecond;
		int i7a = systime.wMilliseconds;

		DWORD dwErr = ::GetLastError();

		for (int i = 0; i < 100; i++)
		{
			wchar_t *szFilepath = NULL;

			if (bTidyMode)
			{
				if (i == 0)
				{
					szFilepath = StringAllocAndFormat(L"%s%s", strFileStart.c_str(), strExtension.c_str());
				}
				else
				{
					int idx = i - 1;
					szFilepath = StringAllocAndFormat(L"%s%d%s", strFileStart.c_str(), idx, strExtension.c_str());

					if (i >= 20)
					{
						bTidyMode = false; // Turn off tidy mode if we're having trouble finding a unique name.
					}
				}
			}
			else
			{
				int i7 = i7a + i;

				szFilepath = StringAllocAndFormat(L"%s%04d%02d%02d%02d%02d%02d%04d%s",
								strFileStart.c_str(), i1,i2,i3,i4,i5,i6,i7, strExtension.c_str());
			}

			if (NULL == szFilepath)
			{
				dwErr = ERROR_NOT_ENOUGH_MEMORY;
				break;
			}
			else
			{
				if (bDontReallyOpen)
				{
					bOpenResult = true;
				}
				else
				{
					funcs.OpenFile(pFileHandle, szFilepath, bReadWrite?(GENERIC_READ|GENERIC_WRITE):GENERIC_WRITE, dwShareMode, CREATE_NEW, dwFlagsAndAttributes);
					dwErr = ::GetLastError();

					if (funcs.IsGoodFileValue(*pFileHandle))
					{
						bOpenResult = true;
					}
				}

				if (bOpenResult)
				{
					if (NULL != pstrFilepathOut)
					{
						*pstrFilepathOut = szFilepath;
					}
					delete [] szFilepath;
					break;
				}
				delete [] szFilepath;

				if (dwErr != ERROR_ALREADY_EXISTS && dwErr != ERROR_FILE_EXISTS) // CreateFile can return either, depending on the CREATE_NEW/CREATE_ALWAYS flag.
				{
					break;
				}
			}
		}

		if (!bOpenResult)
		{
			::SetLastError(dwErr);
			return false;
		}

		return true;
	}

	inline bool OpenTempFileNamePreserveExtension(HANDLE *pFileHandle, const wchar_t *szTempPath, const wchar_t *szPrefix, const wchar_t *szFilenameOrDotExtIn, std::wstring *pstrFilepathOut, bool bDontReallyOpen, DWORD dwShareMode, DWORD dwFlagsAndAttributes, bool bTidyMode, bool bReadWrite)
	{
		FileOpFuncs_Direct fof_direct;
		return LeoHelpers::OpenTempFileNamePreserveExtension(fof_direct, pFileHandle, szTempPath, szPrefix, szFilenameOrDotExtIn, pstrFilepathOut, bDontReallyOpen, dwShareMode, dwFlagsAndAttributes, bTidyMode, bReadWrite);
	}

	// If bDontReallyCreate is true then the function won't really create a directory but will still set *pstrPathOut to a name that is similar to what you'd get normally.
	// On failure *pstrPathOut is guaranteed to be empty.
	// Of course, if you don't create the folder then there is nothing to stop someone else (e.g. another caller of this function)
	// creating it unexpectedly, so the name should only be considered as an example (e.g. to see if we'll generate a path with any non-ASCII chars in it) and never really used.
	// Set bTidyMode when the folder is visible to the user (not hidden in a temp dir) and unlikely to clash with an existing name;
	// in tidy-mode the name will be tried as-is (prefix) and if that already exists a counter is added.
	template< class FileOpFuncs >
	bool CreateTempDirectory(FileOpFuncs &funcs, const wchar_t *szTempPath, const wchar_t *szPrefix, std::wstring *pstrPathOut, bool bDontReallyCreate, bool bTidyMode)
	{
		if (NULL == pstrPathOut)
		{
			return false;
		}

		pstrPathOut->clear();

		BOOL bCreateDirRes = FALSE;

		std::wstring strPathStart;

		if (szTempPath != NULL)
		{
			strPathStart = szTempPath;
		}
		else if (!LeoHelpers::GetTempPath(&strPathStart))
		{
			return false; // last error already set by GetTempPath.
		}

		if (strPathStart.length() > 0 && L'\\' != *(strPathStart.rbegin()) && L'/' != *(strPathStart.rbegin()))
		{
			strPathStart += L'\\';
		}

		if (szPrefix != NULL && szPrefix[0] != L'\0')
		{
			strPathStart += szPrefix;
		}
		else
		{
			strPathStart += L"tmp";
		}

		SYSTEMTIME systime;
		GetSystemTime(&systime);

		int i1 = systime.wYear;
		int i2 = systime.wMonth;
		int i3 = systime.wDay;
		int i4 = systime.wHour;
		int i5 = systime.wMinute;
		int i6 = systime.wSecond;
		int i7a = systime.wMilliseconds;

		DWORD dwErr = ::GetLastError();

		for (int i = 0; i < 100; i++)
		{
			wchar_t *szPath = NULL;

			if (bTidyMode)
			{
				if (i == 0)
				{
					szPath = StringAllocAndFormat(L"%s", strPathStart.c_str());
				}
				else
				{
					int idx = i - 1;
					szPath = StringAllocAndFormat(L"%s%d", strPathStart.c_str(), idx);

					if (i >= 20)
					{
						bTidyMode = false; // Turn off tidy mode if we're having trouble finding a unique name.
					}
				}
			}
			else
			{
				int i7 = i7a + i;

				szPath = StringAllocAndFormat(L"%s%04d%02d%02d%02d%02d%02d%04d",
								strPathStart.c_str(), i1,i2,i3,i4,i5,i6,i7);
			}

			if (NULL == szPath)
			{
				dwErr = ERROR_NOT_ENOUGH_MEMORY;
				break;
			}
			else
			{
				if (bDontReallyCreate)
				{
					bCreateDirRes = TRUE;
				}
				else
				{
					bCreateDirRes = funcs.CreateDirectory(szPath);
					dwErr = ::GetLastError();
				}

				if (bCreateDirRes)
				{
					*pstrPathOut = szPath;
					delete [] szPath;
					break;
				}
				delete [] szPath;

				if (dwErr != ERROR_ALREADY_EXISTS && dwErr != ERROR_FILE_EXISTS)
				{
					break;
				}
			}
		}

		if (!bCreateDirRes)
		{
			::SetLastError(dwErr);
			return false;
		}

		return true;
	}

	inline bool CreateTempDirectory(const wchar_t *szTempPath, const wchar_t *szPrefix, std::wstring *pstrPathOut, bool bDontReallyCreate, bool bTidyMode)
	{
		FileOpFuncs_Direct fof_direct;
		return LeoHelpers::CreateTempDirectory(fof_direct, szTempPath, szPrefix, pstrPathOut, bDontReallyCreate, bTidyMode);
	}

	void GetTimeStampString(std::wstring *pstrOut, bool bDate, bool bTime, bool bSeconds, bool bMilliseconds);

	template< class FileOpFuncs >
	bool RenameUnique(FileOpFuncs &funcs, const wchar_t *szExistingName, const wchar_t *szNewName, bool bPreserveExtension, int iPad, int iMax)
	{
		std::wstring strBaseName = szNewName;
		std::wstring strExt;

		if (bPreserveExtension)
		{
			strExt = LeoHelpers::GetExtensionPart(true, strBaseName.c_str());
			LeoHelpers::TrimExtensionPart(true, &strBaseName);
		}

		typename LeoHelpers::WCharArrayScoper scwc;

		for (int i = 0; i < iMax; ++i)
		{
			scwc.Set(LeoHelpers::StringAllocAndFormat(L"%s_%0*d%s", strBaseName.c_str(), iPad, i, strExt.c_str()));
			if (scwc.IsEmpty())
			{
				::SetLastError(ERROR_NOT_ENOUGH_MEMORY);
				return false;
			}
			if (funcs.MoveFileEx(szExistingName, scwc.Get(), MOVEFILE_COPY_ALLOWED))
			{
				return true;
			}
			if (::GetLastError() != ERROR_ALREADY_EXISTS && ::GetLastError() != ERROR_FILE_EXISTS)
			{
				return false;
			}
		}

		return false;
	}

	bool RenameUnique(const wchar_t *szExistingName, const wchar_t *szNewName, bool bPreserveExtension, int iPad, int iMax);

	bool CopyFile(const wchar_t *szSource, const wchar_t *szDest, HANDLE hAbortEvent, bool bClearReadOnlyAttrib, bool bFailIfExists);
	bool MoveFile(const wchar_t *szSource, const wchar_t *szDest, HANDLE hAbortEvent, bool bFailIfExists);

	// Intended for reading from pipes.
	bool ReadFileUntilDone(HANDLE hFile, LPVOID lpBuffer, DWORD dwNumberOfBytesToRead);

	// Intended for writing to pipes.
	bool WriteFileUntilDone(HANDLE hFile, LPVOID lpBuffer, DWORD dwNumberOfBytesToWrite);

	// Diverts call to FindFirstFileEx with FIND_FIRST_EX_LARGE_FETCH if using wildcards.
	// Falls back on old API if it detects a problem, since some NAS devices mishandle FIND_FIRST_EX_LARGE_FETCH.
	HANDLE FindFirstFileWrapper(LPCWSTR lpFileName, LPWIN32_FIND_DATAW lpFindFileData);

#ifdef SHCreateStreamOnFile
	HRESULT SHCreateStreamOnFile(LPCWSTR pszFile, DWORD grfMode, IStream **ppstm);
#endif

	class FileAndStream
	{
	public:
		// hAbortEvent can be null.
		FileAndStream(const wchar_t *szFilePath, HANDLE hAbortEvent, bool bOpenTempCopy);

		// pStream can be NULL but only if you are never going to call methods which read data or get the steam or file path.
		FileAndStream(const wchar_t *szFileName, IStream *pStream, bool bNoRandomSeek, bool bSlow, bool bSlowSeek);

		FileAndStream(const wchar_t *szFileName, const BYTE *pData, UINT cbDataSize);

		~FileAndStream(); // Warning: Non-virtual destructor.

	private:
		// disallow
		FileAndStream(const FileAndStream &rhs);
		FileAndStream &operator=(const FileAndStream &rhs);

	public:

		// This allows you to set the OpenTempCopy flag after construction.
		// Any subsequent call to GetFilePath will result in the file being copied to a temp path if
		// it had not already been done. Any preceeding call to GetFilePath may, of course, have alredy
		// received the real filename. If the data is in a stream then calling this has no effect as
		// GetFilePath will always cause the data to be saved to a temp file regardless of the flag.
		void SetOpenTempCopy(bool bOpenTemp);

		// The file name may not match the name of the actual file. It is indicative of the original file's name
		// but the file path returned by GetFilePath may be a temp-file with a different name. GetFileName should
		// be used for display purposes and for any logic which depends on the original file's name.
		// Calling GetFileName will not trigger any files or streams to be created.
		bool GetFileName(std::wstring *pstrFileName);
		bool GetFileExtension(std::wstring *pstrFileExtension, bool bIncludeDot);

		// GetNominalFilePath gets a full path that will be similar to, but may not be the same as,
		// the file path returned by GetFilePath. You can use this to test for non-ASCII characters
		// in the file path without triggering a stream to be written to a file as you would if you
		// called GetFilePath.
		bool GetNominalFilePath(std::wstring *pstrNominalFilePath);
		bool GetNominalTempFilePath(std::wstring *pstrNominalTempFilePath);

		bool HasFilePath();
		bool HasStream();

		// These return the current seeking ability. One reason to call these is to test whether or not calling GetStream_NoRef_Conversion_SeekReset will have to do an expensive extraction to create a seekable stream.
		bool HasFastRandomSeek() { return !m_bNoRandomSeekInStream && !m_bSlowStream && !m_bSlowSeekStream; } // Does not imply there is a stream (but it'll be cheap to create if there isn't one).
		bool HasNoRandomSeekStream() { return m_bNoRandomSeekInStream; }
		bool HasSlowStream()         { return m_bSlowStream; }
		bool HasSlowSeekStream()     { return m_bSlowSeekStream; }

		HANDLE GetAbortEvent() { return m_hAbortEvent; } // May return NULL if there isn't one.

		// These calls will cause a file or stream to be created if there isn't one already.
		// Call HasFilePath and HasStream to see what already exists if you can work with both and want
		// to avoid the overhead of conversion.

		bool GetFilePath(std::wstring *pstrFilePath);

		// The stream will not be AddRef'd by this call. It will exist as long as the CFileAndSteam does
		// so you should not normally need to AddRef it.
		// If a stream is returned it will always be fast and seekable.
		// Whenever GetStream_NoRef_Conversion_SeekReset is called the stream's position is reset to the start.
		bool GetStream_NoRef_Conversion_SeekReset(IStream **ppStream, bool *pOutNoRandomSeek, bool *pOutSlow, bool *pOutSlowSeek);

		// Call this to get the current IStream, if any.
		// Unlike the GetStream_NoRef_Conversion_SeekReset() method:
		// - The returned IStream *will* be AddRef'd and you must Release it.
		// - The returned IStream may not be seekable.
		// - If the object contains a filepath that hasn't already been converted into an IStream then that conversion will not happen and the method will fail instead.
		// - The returned IStream will not have its position reset to the start of the IStream (since it may not be seekable).
		bool GetStream_AddRef_NoConversion_NoSeekReset(IStream **ppStream, bool *pOutNoRandomSeek, bool *pOutSlow, bool *pOutSlowSeek);

	private:
		bool PrepareFileForReading();
		bool WriteStreamToTempFile();

	private:
		std::wstring m_strFileName;

		std::wstring m_strFilePath;
		bool m_bDeleteFilePathOnClose;
		bool m_bOpenTempCopyFilePath;

		IStream *m_pStream;
		bool m_bNoRandomSeekInStream;
		bool m_bSlowStream;
		bool m_bSlowSeekStream;

		bool m_bStreamIsFile;

		HANDLE m_hAbortEvent;
	};

	class DirEnum
	{
	private:
		WIN32_FIND_DATA m_wfd;
		HANDLE m_hFind;
		std::wstring m_strDirPath;

		DirEnum(const DirEnum &rhs); // disallow
		DirEnum &operator=(const DirEnum &rhs); // disallow

		bool skipDots();

	public:
		DirEnum() : m_wfd() , m_hFind(INVALID_HANDLE_VALUE) { }
		~DirEnum() { Close(); } // Not virtual.
		bool Init(DWORD *pdwError, const wchar_t *szDirPath, const wchar_t *szPattern);
		void Close();
		bool AtEnd() const { return (m_hFind == INVALID_HANDLE_VALUE); };
		const WIN32_FIND_DATA &CurrentData() const { assert(!AtEnd()); return m_wfd; }
		void CurrentFullPath(std::wstring *pStr) const;
		bool Step(DWORD *pdwError);
	};

	// As well as converting the path to a PIDL for you, this runs SHOpenFolderAndSelectItems on a separate thread so that it does not
	// start dispatching window messages on the calling thread while waiting and potentially messing things up (stupid Windows Shell APIs are stupid).
#ifdef PIDLIST_ABSOLUTE
	bool SHOpenFolderAndSelectItemsWrapper(const wchar_t *szPath);
#endif

	bool GetModuleFileNameToString(HMODULE hModule, std::wstring &str);
};
